
import logging
from typing import TYPE_CHECKING

from patshared import GroupPos, PortMode, PortTypesViewFlag

from .bases.elements import CanvasOptimizeIt
from .bases.group import Group
from .patchcanvas import patchcanvas

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager


_logger = logging.getLogger(__name__)


def change_port_types_view(
        mng: 'PatchbayManager', port_types_view: PortTypesViewFlag,
        force=False):
    if not force and port_types_view is mng.port_types_view:
        return

    ex_ptv = mng.port_types_view
    mng.port_types_view = port_types_view
    _logger.info(
        f"Change Port Types View: {ex_ptv.name} -> {port_types_view.name}")
    # Prevent visual update at each canvas item creation
    # because we may create/remove a lot of ports here

    change_counter = 0

    if len(mng.groups) > 30:
        for group in mng.groups:
            in_outs_ptv = group.ins_ptv | group.outs_ptv

            if in_outs_ptv & port_types_view is not in_outs_ptv & ex_ptv:
                change_counter += 1
                continue

            new_gpos = mng.get_group_position(group.name)
            if group.current_position.needs_redraw(new_gpos):
                change_counter += 1

    if change_counter > 30:
        # Big changes between the current and the next view
        # Strategy is to remove all from canvas and add all what is needed
        # without animation.
        for connection in mng.connections:
            connection.in_canvas = False

        for group in mng.groups:
            group.in_canvas = False
            for portgroup in group.portgroups:
                portgroup.in_canvas = False
            for port in group.ports:
                port.in_canvas = False

        patchcanvas.clear_all()

        with CanvasOptimizeIt(mng):
            for group in mng.groups:
                group.current_position = mng.get_group_position(group.name)
                if (group.outs_ptv | group.ins_ptv) & mng.port_types_view:
                    group.add_to_canvas()
                    group.add_all_ports_to_canvas()

            for connection in mng.connections:
                connection.add_to_canvas()

        patchcanvas.redraw_all_groups()

        mng.sg.port_types_view_changed.emit(mng.port_types_view)
        mng.view().default_port_types_view = mng.port_types_view
        mng.save_view_and_port_types_view()
        return

    rm_all_before = bool(ex_ptv & mng.port_types_view
                            is PortTypesViewFlag.NONE)

    with CanvasOptimizeIt(mng):
        if rm_all_before:
            # there is no common port type between previous and next view,
            # strategy is to remove fastly all contents from the patchcanvas.
            for connection in mng.connections:
                connection.in_canvas = False

            for group in mng.groups:
                group.in_canvas = False
                for portgroup in group.portgroups:
                    portgroup.in_canvas = False
                for port in group.ports:
                    port.in_canvas = False

            patchcanvas.clear_all()

        else:
            for connection in mng.connections:
                connection.remove_from_canvas()

        groups_and_pos = dict[Group, tuple[GroupPos, PortMode, PortMode]]()

        for group in mng.groups:
            new_gpos = mng.get_group_position(group.name)
            in_outs_ptv = group.ins_ptv | group.outs_ptv
            hidden_modes = group.current_position.hidden_port_modes()
            new_hidden_modes = new_gpos.hidden_port_modes()
            redraw_mode = PortMode.NULL

            if (hidden_modes is not new_hidden_modes
                    or mng.port_types_view & in_outs_ptv
                    is not ex_ptv & in_outs_ptv):
                # group needs to be redrawn because visible ports
                # are not the sames.

                if new_hidden_modes is not hidden_modes:
                    # During the animation, we need to see the ports we hide,
                    # so the hidden ports in the new view must be shown.
                    # But, if the ports are hidden in the two views,
                    # we won't show them.
                    for port_mode in PortMode.INPUT, PortMode.OUTPUT:
                        if (not new_hidden_modes & port_mode
                                and hidden_modes & port_mode):
                            redraw_mode |= port_mode

                if (mng.port_types_view & group.ins_ptv
                        is not ex_ptv & group.ins_ptv):
                    redraw_mode |= PortMode.INPUT

                if (mng.port_types_view & group.outs_ptv
                        is not ex_ptv & group.outs_ptv):
                    redraw_mode |= PortMode.OUTPUT

                if not rm_all_before:
                    for portgroup in group.portgroups:
                        if portgroup.port_mode & redraw_mode:
                            portgroup.remove_from_canvas()

                    for port in group.ports:
                        if port.mode & redraw_mode:
                            port.remove_from_canvas()

                if (new_gpos.is_splitted()
                        is not group.current_position.is_splitted()):
                    group.add_to_canvas(gpos=new_gpos)
                else:
                    group.add_to_canvas()

                # only ports which should be hidden in previous and next
                # view will be hidden (before to animate).
                for port in group.ports:
                    port.add_to_canvas(
                        ignore_gpos=True,
                        hidden_sides=hidden_modes & new_hidden_modes)

                for portgroup in group.portgroups:
                    portgroup.add_to_canvas()

            for port in group.ports:
                if port.in_canvas:
                    new_gpos.has_sure_existence = True
                    break
            else:
                group.remove_from_canvas()

            restore_mode = PortMode.NULL
            for pmode in (PortMode.INPUT, PortMode.OUTPUT):
                if (group.current_position.hidden_port_modes() & pmode
                        and not new_gpos.hidden_port_modes() & pmode):
                    restore_mode |= pmode
            groups_and_pos[group] = (new_gpos, redraw_mode, restore_mode)

        for conn in mng.connections:
            conn.add_to_canvas()

    if groups_and_pos:
        patchcanvas.canvas.scene.prevent_box_user_move = True

        for group, gpos_redraw in groups_and_pos.items():
            group.set_group_position(*gpos_redraw)

        patchcanvas.repulse_all_boxes()

    mng.sg.port_types_view_changed.emit(mng.port_types_view)
    mng.view().default_port_types_view = mng.port_types_view
    mng.save_view_and_port_types_view()

def new_view(mng: 'PatchbayManager', view_number: int | None =None,
             exclusive_with: dict[int, PortMode] | None =None):
    '''create a new view and switch directly to this view.

    If `view_number` is not set, it will choose the first available
    number.

    if `exclusive_with` is set, all non matching boxes will be hidden,
    and new view will be a white list view.'''

    new_num = mng.views.add_view(
        view_number, default_ptv=mng.port_types_view)
    if new_num is None:
        _logger.warning(f'failed to create new view n°{view_number}')
        return

    if exclusive_with is None:
        for gpos in mng.views.iter_group_poses(view_num=mng.view_number):
            mng.views.add_group_pos(new_num, gpos.copy())

        if mng.view().is_white_list:
            mng.views[new_num].is_white_list = True

    else:
        mng.views[new_num].is_white_list = True

        v = mng.view().ptvs[mng.port_types_view]
        mng.views[new_num].ptvs[mng.port_types_view] = new_v = {}

        for group_id, port_mode in exclusive_with.items():
            group = mng.get_group_from_id(group_id)
            if group is not None:
                new_v[group.name] = gpos = v[group.name].copy()
                for pmode, box_pos in gpos.boxes.items():
                    if not port_mode & pmode:
                        box_pos.set_hidden(True)

    mng.view().default_port_types_view = mng.port_types_view
    mng.view_number = new_num
    mng.set_views_changed()
    change_port_types_view(mng, mng.port_types_view, force=True)

def rename_current_view(mng: 'PatchbayManager', new_name: str):
    mng.view().name = new_name
    mng.set_views_changed()

def change_view(mng: 'PatchbayManager', view_number: int):
    if not view_number in mng.views.keys():
        mng.new_view(view_number=view_number)
        return

    mng.view_number = view_number
    new_view = mng.views.get(mng.view_number)
    if new_view is None:
        ptv = mng.port_types_view
    else:
        ptv = new_view.default_port_types_view

    change_port_types_view(mng, ptv, force=True)
    mng.sg.view_changed.emit(view_number)

def remove_view(mng: 'PatchbayManager', view_number: int):
    if len(mng.views) <= 1:
        _logger.error(
            f"Will not remove view {view_number}, "
            "to ensure there is at least one view.")
        return

    rm_current_view = bool(view_number is mng.view_number)
    mng.views.remove_view(view_number)

    if rm_current_view:
        switch_to_view = -1
        for view_num in mng.views.keys():
            if view_num < view_number:
                switch_to_view = view_num
            elif switch_to_view == -1:
                switch_to_view = view_num
                break

        mng.change_view(switch_to_view)

    mng.set_views_changed()

def clear_absents_in_view(mng: 'PatchbayManager', only_current_ptv=False):
    if only_current_ptv:
        mng.views.clear_absents(
            mng.view_number, mng.port_types_view,
            set([g.name for g in mng.groups
                    if g.is_in_port_types_view(mng.port_types_view)]))
        return

    for ptv in mng.view().ptvs.keys():
        mng.views.clear_absents(
            mng.view_number, ptv,
            set([g.name for g in mng.groups
                    if g.is_in_port_types_view(ptv)]))

def remove_all_other_views(mng: 'PatchbayManager'):
    view_nums = [n for n in mng.views.keys() if n != mng.view_number]
    for n in view_nums:
        mng.views.remove_view(n)
    mng.set_views_changed()

def change_view_number(mng: 'PatchbayManager', new_num: int):
    if new_num == mng.view_number:
        return

    mng.views.change_view_number(mng.view_number, new_num)
    mng.view_number = new_num
    mng.sg.views_changed.emit()

def write_view_data(
        mng: 'PatchbayManager', view_number: int, name: str | None =None,
        port_types: PortTypesViewFlag | None =None,
        white_list_view=False):
    view_data = mng.views.get(view_number)
    if view_data is None:
        if port_types is None:
            port_types = PortTypesViewFlag.ALL

        mng.views.add_view(view_number, port_types)
        view_data = mng.views[view_number]

    if name is not None:
        view_data.name = name

    if port_types is not None:
        view_data.default_port_types_view = port_types

    view_data.is_white_list = white_list_view

    mng.set_views_changed()



import configparser
import logging
from pathlib import Path
import sys

from qtpy.QtGui import QColor

sys.path.insert(0, str(Path(__file__).parents[1]))

from patchbay.patchcanvas import theme, theme_manager

from color import Color, compare_colors

_logger = logging.getLogger(__name__)


def qcolor_name(qcolor: QColor) -> str:
    if qcolor.alphaF() == 1.0:
        return qcolor.name()

    alphaf = '%.2f' % qcolor.alphaF()
    while alphaf.endswith('0'):
        alphaf = alphaf[:-1]
        if alphaf[-1] == '.':
            alphaf += '0'
            break

    return f'{qcolor.name()} ** {alphaf}'

def tuple_key(input: tuple[str, str]) -> str:
    return f'[{input[0]}]{input[1]}'

def read_theme(theme_path: Path):
    # tm = theme_manager.ThemeManager(
    #     (Path(__file__).parents[2] / 'themes',), outside=True)
    # tm.set_theme(theme_name)

    conf = configparser.ConfigParser()
    try:
        # we don't need the file_list
        # it is just a convenience to mute conf.read
        file_list = conf.read(theme_path)
    except configparser.DuplicateOptionError as e:
        _logger.error(str(e))
        return
    except:
        _logger.error(f"failed to open {theme_path}")
        return

    thd = theme_manager.ThemeManager._convert_configparser_object_to_dict(conf)
    thdn = {}
    for key, value in thd.items():
        thdn[key] = {}
        for kkey, vvalue in value.items():
            thdn[key][kkey] = vvalue
    
    th = theme.Theme()
    th.read_theme(thd, theme_path, for_linter=True)    
    aliases = dict[str, str]()
    colors = dict[str, Color]()

    rwt_aliases: dict = thdn.get('aliases', {})
    for section_name, section in conf.items():
        if section_name != 'aliases':
            continue
        
        for key, value in section.items():
            qcolor = theme.to_qcolor(value)
            if qcolor is None:
                _logger.warning(f'alias "{key}" is not a color')
                continue
            aliases[key] = qcolor_name(qcolor)
            rwt_aliases[key] = qcolor_name(qcolor)

    print(aliases)
    aliases_colors = dict[str, set[str]]()
    for alias, color_name in aliases.items():
        exst = aliases_colors.get(color_name)
        if exst is None:
            aliases_colors[color_name] = {alias}
        else:
            aliases_colors[color_name].add(alias)
        
        color = colors.get(color_name)
        if color is None:
            color = Color(color_name)
            colors[color_name] = color

    for section_name, section in conf.items():
        if section_name in ('Theme', 'aliases'):
            continue
        
        for key, value in section.items():
            if not (key.endswith('-color')
                    or key in ('background', 'background2')):
                continue

            for word in value.split():
                if word in aliases:
                    _logger.debug(
                        f"Ignored word '{word}' in [{section_name}]{section} : "
                        f"it is an alias")
                    break
                
                if '*' in word:
                    continue

                try:
                    float(word)
                except:
                    pass
                else:
                    continue

                try:
                    qcolor_ = QColor(word)
                    assert qcolor_.isValid()
                except:
                    _logger.info(f"ignore  '{word}', not a color")
                    continue
                
                color_name = qcolor_.name()
                
                if word.startswith(('#', '-#', 'rgb(', 'rgba(')):
                    if qcolor_.alphaF() != 1.0:
                        new_col_name = qcolor_name(qcolor_)
                        _logger.warning(
                            f'color {word} has an alpha channel, '
                            f'prefer to use {new_col_name}')
                        thdn[section_name][key] = value.replace(word, new_col_name)
                    color = colors.get(color_name)
                    if color is None:
                        color = Color(color_name)
                        colors[color_name] = color
                        
                    color.exists_in.add(f'[{section_name}]{key}')
    
    for color_name, color in colors.items():
        if not color.exists_in:
            continue

        multi_uses = '\n    '.join(color.exists_in)

        if color_name in aliases_colors.keys():
            multi_alias = '|'.join(aliases_colors[color_name])
            _logger.warning(
                f'{color_name} is used in\n    {multi_uses}\n'
                f'        could be replaced by alias {multi_alias}')
        elif len(color.exists_in) >= 2:
            _logger.warning(
                f'{color_name} is used in\n    {multi_uses}\n'
                '        It would be better to choose an alias')
    
    compare_colors(colors)

    # print('--- USED COLORS ---')
    # sorted_colors = sorted(colors.keys())
    # for scol in sorted_colors:
    #     if scol in aliases_colors.keys():
    #         print(scol, '|'.join(aliases_colors[scol]))
    #     else:
    #         print(scol)
        
    #     for oth, ratio in colors[scol].equivalents:
    #         ratio_str = '%.2f' % ratio
    #         aliases = aliases_colors.get(oth)
    #         if aliases is None:
    #             print(f'    = {oth} * {ratio_str}')
    #         else:
    #             for alias_ in aliases:
    #                 print(f'    = {alias_} * {ratio_str}')            

    for thchild in th.all_childs():
        for attr in theme.theme._DEFAULT_STYLE_ATTRS.keys():
            value = thchild._attrs.get(attr)
            if value is None:
                continue
            if thchild._parent.get_value_of(attr) == value: # type:ignore
                _logger.warning(f'[{thchild._path}]{attr} already defined')

    return thdn

def write_theme(thd: dict, theme_path: Path):
    conf = configparser.ConfigParser()
    out_path = theme_path.parent / 'theme.new.conf'

    for key, value in thd.items():
        conf[key] = value

    try:
        with open(out_path, 'w') as f:
            conf.write(f)
    except BaseException as e:
        _logger.error(f'Failed to write {out_path}\n{str(e)}')
    else:
        _logger.warning(f'File written: {out_path}')


if __name__ == '__main__':
    if len(sys.argv) < 2:
        _logger.error("Put a theme file as argument, please !")
        sys.exit(1)

    theme_name = sys.argv[1]
    theme_path = Path(__file__).parents[2] / 'themes' / theme_name / 'theme.conf'
    
    thd = read_theme(theme_path)
    write_theme(thd, theme_path) # type:ignore


import time
from typing import TYPE_CHECKING

from qtpy.QtCore import QPointF, QRectF

from patshared import PortMode

from ..box_widget import BoxWidget
from ..grouped_lines_widget import GroupedLinesWidget
from ..init_values import canvas, AliasingReason, BoxHidding, Joining

from .scene_utils import MovingBox

if TYPE_CHECKING:
    from .scene import PatchScene


MOVE_DURATION = 0.300 # 300ms
MOVE_TIMER_INTERVAL = 20 # 20 ms step animation (50 Hz)
MOVE_N_STEPS = (MOVE_DURATION * 1000) / MOVE_TIMER_INTERVAL


class _MoveTime:
    started_at = 0.0
    last_time = 0.0


def _start_move_timer(scene: 'PatchScene'):
    if scene._move_box_timer.isActive():
        return

    _MoveTime.started_at = time.time()
    _MoveTime.last_time = _MoveTime.started_at
    scene._move_box_timer.start()

def move_boxes_animation(scene: 'PatchScene'):
    # Animation is nice but not the priority.
    # Do not ensure all steps are played
    # but just move the box where it has to go now.
    move_time = time.time()
    time_since_start = move_time - _MoveTime.started_at
    ratio = min(1.0, time_since_start / MOVE_DURATION)

    if _MoveTime.last_time == _MoveTime.started_at:
        # this is the first animation step
        if time_since_start > 0.33 * MOVE_DURATION:
            # this seems to be a big patch,
            # animation won't be pretty anyway,
            # let's finish it now.
            ratio = 1.0
    elif move_time - _MoveTime.last_time > 0.002 * MOVE_TIMER_INTERVAL:
        # this is not the first animation step.
        # The timer called this method two times too late,
        # i.e. >40ms instead of 20ms after the previous step,
        # anti-aliasing is de-activated for a smoother animation.
        # It will be re-activated at animation end.
        canvas.set_aliasing_reason(AliasingReason.ANIMATION, True)

    _MoveTime.last_time = move_time

    lws = set[GroupedLinesWidget]()
    zoom_ratio = scene.zoom_ratio
    useful = False
    
    for box, moving_box in scene.move_boxes.items():
        if not useful:
            useful = moving_box.is_useful

        if moving_box.needs_move:
            max_range = moving_box.max_distance * zoom_ratio
            if ratio >= 1.0 or max_range == 0:
                box.set_top_left((moving_box.to_pt.x(), moving_box.to_pt.y()))
            else:
                if max_range < MOVE_N_STEPS:
                    tmp_ratio = min(1.0, ratio * MOVE_N_STEPS / max_range)
                else:
                    tmp_ratio = ratio
                
                x = (moving_box.from_pt.x()
                     + ((moving_box.to_pt.x() - moving_box.from_pt.x())
                        * (tmp_ratio ** 0.6)))

                y = (moving_box.from_pt.y()
                     + ((moving_box.to_pt.y() - moving_box.from_pt.y())
                        * (tmp_ratio ** 0.6)))
                
                if tmp_ratio >= 1.0:
                    box.set_top_left(
                        (moving_box.to_pt.x(), moving_box.to_pt.y()))
                    moving_box.needs_move = False
                else:
                    box.set_top_left(
                        (round(x) * zoom_ratio / zoom_ratio,
                         round(y) * zoom_ratio / zoom_ratio))
            box.repaint_lines(fast_move=True)

        if moving_box.is_wrapping:
            box.animate_wrapping(ratio)

        if moving_box.hidding_state in (BoxHidding.HIDDING,
                                        BoxHidding.RESTORING):
            if moving_box.hidding_state is BoxHidding.HIDDING:
                box.animate_hidding(ratio)
            else:
                box.animate_restoring(ratio)

            for lw in GroupedLinesWidget.widgets_for_box(
                    box._group_id, box._port_mode):
                if lw not in lws:
                    lw.animate_hidding(ratio)
                    lws.add(lw)

    if not useful:
        # stop animation now if all moving boxes have no change to make
        ratio = 1.0

    scene.resize_the_scene(ratio)

    if ratio >= 1.0:
        # Animation is finished
        scene._move_box_timer.stop()
        canvas.set_aliasing_reason(AliasingReason.ANIMATION, False)
        scene.prevent_box_user_move = False
        GroupedLinesWidget.animation_finished()

        # box update positions is forbidden while widget 
        # is in self.move_boxes,
        # so we copy the list before to clear it,
        # then we can ask update_positions on widgets
        boxes = [
            b for b, mb in scene.move_boxes.items()
            if not (mb.is_joining or mb.hidding_state is BoxHidding.HIDDING)]

        scene.move_boxes.clear()

        for box in boxes:
            if box.update_positions_pending:
                box.update_positions()
            else:
                box.update_ports()

        canvas.qobject.move_boxes_finished.emit()

def add_box_to_animation(
        scene: 'PatchScene', box_widget: BoxWidget, to_x: int, to_y: int,
        joining=Joining.NO_CHANGE, joined_rect=QRectF()):
    '''add a box to the move animation, to_x and to_y refer
    to the top left of the box at the end of animation.
    if joining is set to Joining.YES, joined_rect must be set'''

    moving_box = scene.move_boxes.get(box_widget)
    if moving_box is None:
        # box is not already moving, create a MovingBox instance
        moving_box = MovingBox(box_widget)
        if joining is Joining.YES:
            moving_box.is_joining = True
        scene.move_boxes[box_widget] = moving_box
    else:
        # box is already moving, check joining state and change it
        # if needed.
        if moving_box.is_joining and joining is Joining.NO:
            moving_box.is_joining = False
            canvas.qobject.rm_group_to_join(box_widget._group_id)

    moving_box.from_pt = QPointF(*box_widget.top_left())
    moving_box.to_pt = QPointF(to_x, to_y)
    moving_box.needs_move = bool(moving_box.from_pt != moving_box.to_pt)

    if joining is Joining.YES or not box_widget.isVisible():
        moving_box.final_rect = joined_rect

    elif joining is Joining.NO_CHANGE and moving_box.is_joining:
        final_rect = QRectF(
            0.0, 0.0,
            moving_box.final_rect.width(),
            moving_box.final_rect.height())
        final_rect.translate(moving_box.to_pt)
        moving_box.final_rect = final_rect

    elif not moving_box.is_joining:
        aft_wrap_rect = box_widget.after_wrap_rect()
        final_rect = QRectF(
            0.0, 0.0, aft_wrap_rect.width(), aft_wrap_rect.height())
        final_rect.translate(moving_box.to_pt)
        moving_box.final_rect = final_rect

    else:
        # can not happens
        # would means moving_box.is_joining and joining is JOINING.NO,
        # It is prevented
        moving_box.final_rect = joined_rect

    if joining is not Joining.NO_CHANGE:
        moving_box.is_joining = True if joining is Joining.YES else False

    # save the group position
    group = canvas.get_group(box_widget._group_id)
    if group is not None:
        if moving_box.is_joining:
            if not joined_rect.isNull():
                group.gpos.boxes[PortMode.BOTH].pos = (to_x, to_y)
        else:
            group.gpos.boxes[box_widget._port_mode].pos = (to_x, to_y)
        canvas.cb.group_pos_modified(group.group_id)

    _start_move_timer(scene)

    if canvas.aliasing_reason:
        # if antialiasing is already prevented
        # we need to keep it prevented at animation start
        canvas.set_aliasing_reason(AliasingReason.ANIMATION, True)

def remove_box_from_animation(scene: 'PatchScene', box_widget: BoxWidget):
    if scene.prevent_box_user_move:
        # should not happens.
        # For now we can remove a box from animation
        # only by moving box manually,
        # and this is prevented by this attr in box_widget_moth.
        return

    if box_widget in scene.move_boxes:
        scene.move_boxes.pop(box_widget)

def add_box_to_animation_wrapping(
        scene: 'PatchScene', box_widget: BoxWidget, wrap: bool):
    moving_box = scene.move_boxes.get(box_widget)
    if moving_box is None:
        moving_box = MovingBox(box_widget)
        scene.move_boxes[box_widget] = moving_box

    aft_wrap_rect = box_widget.after_wrap_rect()
    final_rect = QRectF(0.0, 0.0, aft_wrap_rect.width(), aft_wrap_rect.height())
    moving_box.final_rect = \
        final_rect.translated(moving_box.to_pt)
    moving_box.is_wrapping = True

    _start_move_timer(scene)

def add_box_to_animation_hidding(scene: 'PatchScene', box_widget: BoxWidget):
    moving_box = scene.move_boxes.get(box_widget)
    if moving_box is None:
        moving_box = MovingBox(box_widget)
        scene.move_boxes[box_widget] = moving_box

    moving_box.final_rect = QRectF()
    moving_box.hidding_state = BoxHidding.HIDDING

    for port_mode in PortMode.OUTPUT, PortMode.INPUT:
        if port_mode not in box_widget.port_mode:
            continue

        for lw in GroupedLinesWidget.widgets_for_box(
                box_widget._group_id, port_mode):
            lw.set_mode_hidding(port_mode, BoxHidding.HIDDING)

    _start_move_timer(scene)

def add_box_to_animation_restore(scene: 'PatchScene', box_widget: BoxWidget):
    moving_box = scene.move_boxes.get(box_widget)
    if moving_box is None:
        moving_box = MovingBox(box_widget)
        scene.move_boxes[box_widget] = moving_box

    moving_box.from_pt = moving_box.to_pt

    aft_wrap_rect = box_widget.after_wrap_rect()
    final_rect = QRectF(
        0.0, 0.0, aft_wrap_rect.width(), aft_wrap_rect.height())
    moving_box.final_rect = \
        final_rect.translated(moving_box.to_pt)

    if moving_box.hidding_state is BoxHidding.NONE:
        box_widget.animate_restoring(0.0)
    moving_box.hidding_state = BoxHidding.RESTORING

    for port_mode in PortMode.OUTPUT, PortMode.INPUT:
        if port_mode not in box_widget._port_mode:
            continue

        for lw in GroupedLinesWidget.widgets_for_box(
                box_widget._group_id, port_mode):
            lw.set_mode_hidding(port_mode, BoxHidding.RESTORING)

    _start_move_timer(scene)



from functools import cached_property
import logging
from typing import Iterator, TYPE_CHECKING, Any

from qtpy import QT5
from qtpy.QtCore import Qt
from qtpy.QtGui import (
    QBrush, QImage, QFontDatabase, QFont, QPen, QColor, QFontMetricsF)

from . import theme_cache
from .theme_utils import to_qcolor, rail_float, rail_int, ThemeFile
from .theme_structs import Margin, BorderMode, Align

if TYPE_CHECKING:
    from .style_attributers import UslStyleAttributer


_logger = logging.getLogger(__name__)


class StyleAttributer:
    def __init__(self, path: str,
                 parent: 'StyleAttributer | None' =None):
        self.subs = list[str]()

        self._attrs: dict[str, Any] = {}
 
        self._path = path
        self._parent = parent

        if TYPE_CHECKING:
            assert isinstance(self._parent, StyleAttributer)

    def child(self, path: str) -> 'StyleAttributer | None':
        begin, _, end = path.partition('.')
        if begin not in self.subs:
            return None
        
        child_: StyleAttributer = self.__getattribute__(begin)
        if not end:
            return child_
        
        return child_.child(end)

    def childs(self) -> 'Iterator[StyleAttributer]':
        for sub in self.subs:
            yield self.__getattribute__(sub)

    def all_childs(self) -> 'Iterator[StyleAttributer]':
        for sub in self.subs:
            child = self.__getattribute__(sub)
            yield child

            if TYPE_CHECKING:
                assert isinstance(child, StyleAttributer)

            for sub_ in child.all_childs():
                yield sub_

    def inherit(self, other: 'StyleAttributer'):
        for key, value in other._attrs.items():
            self._attrs[key] = value

        for sub_ in self.subs:
            if sub_ in other.subs:
                self.child(sub_).inherit(other.child(sub_)) # type:ignore

    @property
    def log_path(self):
        return f'[{self._path[1:]}]'

    def set_attribute(self, attribute: str, value: str | float):
        err = False
        match attribute:
            case 'inherits':
                pass
            case 'border-color'|'text-color'|'background'|'background2':
                if isinstance(value, str):
                    self._attrs[attribute] = to_qcolor(value)
                if self._attrs.get(attribute) is None:
                    err = True
        
            case 'background-image':
                image_path = (
                    ThemeFile.path.parent / 'images' / str(value))
                image = None
                if image_path.is_file():
                    image = QImage(str(image_path))
                    image.setDevicePixelRatio(3.0)
                    if image.isNull():
                        _logger.error(
                            f'{self._path}:{attribute} : '
                            f'Failed to load {image_path} as an image')
                        image = None
                else:
                    _logger.error(
                        f"{self.log_path}{attribute} can not find image at {image_path}")
                self._attrs[attribute] = image

            case 'border-width'|'border-radius'|'font-size'| \
                    'port-offset'|'port-in-offset'|'port-out-offset'| \
                    'port-spacing'|'port-type-spacing'| \
                    'icon-size'|'grid-min-width'|'grid-min-height'| \
                    'margin'|'margin-top'|'margin-bottom'|'margin-sides'|\
                    'margin-ports-side'|'margin-free-side'|'margin-top-side':
                if isinstance(value, (int, float)):
                    match attribute:
                        case 'border-width':
                            min_, max_ = 0, 20
                        case 'border-radius':
                            min_, max_ = 0, 50
                        case 'margin'|'margin-top'|'margin-bottom'|\
                                'margin-sides'|'margin-ports-side'|\
                                'margin-free-side'|'margin-top-side':
                            min_, max_ = -50, 50
                        case 'font-size':
                            min_, max_ = 1, 200
                        case 'port-spacing'|'port-type-spacing':
                            min_, max_ = 0, 100
                        case 'icon-size':
                            min_, max_ = 8, 1024
                        case 'grid-min-width'|'grid-min-height':
                            min_, max_ = 1, 100000
                        case _:
                            min_, max_ = -20, 20
                    self._attrs[attribute] = rail_float(value, min_, max_)

                    match attribute:
                        # theses attributes redefine others
                        case 'port-offset':
                            self._attrs['port-in-offset'] = \
                                self._attrs['port-out-offset'] = \
                                    self._attrs[attribute]
                        case 'margin':
                            for key in ('top', 'bottom', 'sides',
                                        'ports-side', 'free-side', 'top-side'):
                                self._attrs[f'margin-{key}'] = \
                                    self._attrs[attribute]
                else:
                    err = True
            
            case 'border-style':
                if isinstance(value, str):
                    value = value.lower()
                    border_style: Qt.PenStyle | None

                    match value:
                        case 'solid'|'normal':
                            border_style = Qt.PenStyle.SolidLine
                        case 'nopen'|'none':
                            border_style = Qt.PenStyle.NoPen
                        case 'dash':
                            border_style = Qt.PenStyle.DashLine
                        case 'dashdot':
                            border_style = Qt.PenStyle.DashDotLine
                        case 'dashdotdot':
                            border_style = Qt.PenStyle.DashDotDotLine
                        case _:
                            border_style = None
                            
                    if border_style is None:
                        err = True
                    else:
                        self._attrs[attribute] = border_style
                else:
                    err = True
                    
            case 'font-name':
                if isinstance(value, str):
                    self._attrs[attribute] = value

                    # add the font to database if it is an embedded font
                    for ext in ('ttf', 'otf'):
                        embedded_path = (
                            ThemeFile.path.parent
                            / 'fonts' / f"{value}.{ext}")
                        if embedded_path.is_file():
                            QFontDatabase.addApplicationFont(
                                str(embedded_path))
                            break
                else:
                    err = True
                    
            case 'font-weight':
                if isinstance(value, str):
                    match value.lower():
                        case 'thin':
                            self._attrs[attribute] = QFont.Weight.Thin
                        case 'extralight'|'extra-light'|'extra_light':
                            self._attrs[attribute] = QFont.Weight.ExtraLight
                        case 'light':
                            self._attrs[attribute] = QFont.Weight.Light
                        case 'normal':
                            self._attrs[attribute] = QFont.Weight.Normal
                        case 'medium':
                            self._attrs[attribute] = QFont.Weight.Medium
                        case 'demibold'|'demi-bold'|'demi_bold':
                            self._attrs[attribute] = QFont.Weight.DemiBold
                        case 'bold':
                            self._attrs[attribute] = QFont.Weight.Bold
                        case 'extrabold'|'extra-bold'|'extra_bold':
                            self._attrs[attribute] = QFont.Weight.ExtraBold
                        case 'black':
                            self._attrs[attribute] = QFont.Weight.Black
                        case _:
                            err = True

                elif isinstance(value, (int, float)):
                    weight = rail_int(value, 1, 1000)
                    if QT5:
                        self._attrs[attribute] = (weight - 1)  // 10
                    else:
                        self._attrs[attribute] = weight

                else:
                    err = True
            
            case 'border-mode':
                if isinstance(value, str):
                    match value.lower():
                        case 'minimal':
                            self._attrs[attribute] = BorderMode.MINIMAL
                        case 'sides':
                            self._attrs[attribute] = BorderMode.SIDES
                        case _:
                            self._attrs[attribute] = BorderMode.DEFAULT
                else:
                    err = True
            
            case 'output-align':
                if isinstance(value, str):
                    match value.lower():
                        case 'right':
                            self._attrs[attribute] = Align.RIGHT
                        case _:
                            self._attrs[attribute] = Align.LEFT
                else:
                    err = True
            
            case 'port-offset-mode'|\
                    'port-in-offset-mode'|'port-out-offset-mode':
                if isinstance(value, str):
                    self._attrs[attribute] = value

                    if attribute == 'port-offset-mode':
                        self._attrs['port-in-offset-mode'] = \
                            self._attrs['port-out-offset-mode'] = \
                                self._attrs['port-offset-mode']
                else:
                    err = True
            
            case 'visible'|'drilled':
                if isinstance(value, str):
                    hcb = value.lower() not in ('false', 'no')
                elif isinstance(value, (int, float)):
                    hcb = bool(value)
                else:
                    _logger.error(f'[{self._path}]{attribute} : '
                                  f'"{value}" is not a valid value')
                    return
                
                self._attrs[attribute] = hcb
            
            case _:
                _logger.error(f"{self.log_path}{attribute} unknown key !")

        if err:
            _logger.error(
                f"{self._path}: invalid value for {attribute}: {str(value)}")

    def set_style_dict(self, context: str, style_dict: dict):
        if context:
            begin, point, end = context.partition('.')

            if begin not in self.subs:
                _logger.error(f"{self._path}: invalid ignored key: {begin}")
                return
            
            self.child(begin).set_style_dict(end, style_dict) # type:ignore
            return

        for key, value in style_dict.items():
            self.set_attribute(key, value)

    def get_value_of(self, attribute: str, orig_path='', needed_attribute=''):
        '''return the value of given attribute for this theme section.
        if this value is not present in this theme section,
        it will look into parent sections.
        Note that for 'selected' section, it will look in 'selected' section
        of parent before looking in parent section.'''
        if not orig_path:
            orig_path = self._path

        for path_end in ('selected',):
            if TYPE_CHECKING:
                assert isinstance(self, UslStyleAttributer)

            if (orig_path.endswith('.' + path_end)
                    and path_end in self.subs
                    and self._path + '.' + path_end != orig_path):
                sel_attr = self.selected._attrs.get(attribute)
                if sel_attr is None:
                    if needed_attribute:
                        if self.selected._attrs.get(needed_attribute) is None:
                            return None
                    continue
                return sel_attr                        

        if self._attrs.get(attribute) is None:
            if (needed_attribute
                    and self._attrs.get(needed_attribute) is not None):
                return None

            if self._parent is None:
                _logger.error(
                    f"get value of: {self._path} None value and no parent")
                return None

            return self._parent.get_value_of(
                attribute, orig_path, needed_attribute)

        return self._attrs.get(attribute)

    @cached_property
    def fill_pen(self) -> QPen:
        return QPen(
            QBrush(self.get_value_of('border-color')),
            self.get_value_of('border-width'),
            self.get_value_of('border-style'))

    @cached_property
    def border_radius(self) -> float:
        return self.get_value_of('border-radius') # type:ignore

    @cached_property
    def background_color(self) -> QColor:
        return self.get_value_of('background') # type:ignore

    @cached_property
    def background2_color(self) -> QColor | None:
        return self.get_value_of('background2', # type:ignore
                                 needed_attribute='background')

    @cached_property
    def background_image(self) -> QImage:
        return self.get_value_of('background-image') # type:ignore

    @cached_property
    def margin(self) -> Margin:
        margin = Margin()
        margin.top = self.get_value_of('margin-top') # type:ignore
        margin.bottom = self.get_value_of('margin-bottom') # type:ignore
        margin.sides = self.get_value_of('margin-sides') # type:ignore
        margin.ports_side = self.get_value_of('margin-ports-side') # type:ignore
        margin.free_side = self.get_value_of('margin-free-side') # type:ignore
        margin.top_side = self.get_value_of('margin-top-side') # type:ignore
        return margin

    @cached_property
    def padding(self) -> Margin:
        padding = Margin()
        padding.top = self.get_value_of('padding-top') # type:ignore
        padding.bottom = self.get_value_of('padding-bottom') # type:ignore
        padding.sides = self.get_value_of('padding-sides') # type:ignore
        padding.ports_side = self.get_value_of('padding-ports-side') # type:ignore
        padding.free_side = self.get_value_of('padding-free-side') # type:ignore
        padding.top_side = self.get_value_of('padding-top-side') # type:ignore
        return padding

    @cached_property
    def margin_empty(self) -> Margin:
        return Margin()

    @cached_property
    def text_color(self) -> QColor:
        return self.get_value_of('text-color') # type:ignore

    @cached_property
    def font(self) -> QFont:
        font = QFont(self.get_value_of('font-name'))
        font.setPixelSize(int(self.get_value_of('font-size'))) # type:ignore
        font.setWeight(int(self.get_value_of('font-weight'))) # type:ignore
        return font

    @cached_property
    def border_mode(self) -> BorderMode:
        return self.get_value_of('border-mode') # type:ignore

    @cached_property
    def border_width(self) -> float:
        '''The border width defined in theme,
        or 0.0 if there is no border'''
        if self.get_value_of('border-style') == Qt.PenStyle.NoPen:
            return 0.0
        return self.get_value_of('border-width') # type:ignore

    @cached_property
    def output_align(self) -> Align:
        return self.get_value_of('output-align') # type:ignore

    @cached_property
    def port_in_offset(self) -> float:
        return self.get_value_of('port-in-offset') # type:ignore

    @cached_property
    def port_out_offset(self) -> float:
        return self.get_value_of('port-out-offset') # type:ignore

    @cached_property
    def port_in_offset_mode(self) -> str:
        return self.get_value_of('port-in-offset-mode') # type:ignore

    @cached_property
    def port_out_offset_mode(self) -> str:
        return self.get_value_of('port-out-offset-mode') # type:ignore

    @cached_property
    def port_spacing(self) -> float:
        return self.get_value_of('port-spacing') # type:ignore

    @cached_property
    def port_type_spacing(self) -> float:
        return self.get_value_of('port-type-spacing') # type:ignore

    @cached_property
    def icon_size(self) -> float:
        return self.get_value_of('icon-size') # type:ignore

    @cached_property
    def grid_min_width(self) -> float:
        return self.get_value_of('grid-min-width') # type:ignore

    @cached_property
    def grid_min_height(self) -> float:
        return self.get_value_of('grid-min-height') # type:ignore

    @cached_property
    def visible(self) -> bool:
        return self.get_value_of('visible') # type:ignore

    @cached_property
    def drilled(self) -> bool:
        return self.get_value_of('drilled') # type:ignore

    @cached_property
    def _titles_templates_cache(self) -> theme_cache.TitleCache:
        font_name = str(self.get_value_of('font-name'))
        font_size = str(self.get_value_of('font-size'))
        font_width = str(self.get_value_of('font-weight'))

        return theme_cache.get_title_templates_cache(
            font_name, font_size, font_width)

    @cached_property
    def _font_metrics_cache(self) -> dict[str, float]:
        font_name = str(self.get_value_of('font-name'))
        font_size = str(self.get_value_of('font-size'))
        font_width = str(self.get_value_of('font-weight'))

        return theme_cache.get_font_metrics_cache(
            font_name, font_size, font_width)

    def get_text_width(self, string: str) -> float:
        if string in self._font_metrics_cache.keys():
            return self._font_metrics_cache[string]

        tot_size = 0.0
        for s in string:
            if s in self._font_metrics_cache.keys():
                tot_size += self._font_metrics_cache[s]
            else:
                letter_size = QFontMetricsF(self.font).horizontalAdvance(s)
                self._font_metrics_cache[s] = letter_size
                tot_size += letter_size

        self._font_metrics_cache[string] = tot_size

        return tot_size
    
    def save_title_templates(
            self, title: str, icon_size: int, templates: list):
        if not title in self._titles_templates_cache:
            self._titles_templates_cache[title] = {}

        self._titles_templates_cache[title][icon_size] = templates

    def get_title_templates(
            self, title: str, icon_size: int) -> list[dict[str, int]]:
        if (title in self._titles_templates_cache
                and icon_size in self._titles_templates_cache[title]):
            return self._titles_templates_cache[title][icon_size]
        return []




import logging
from typing import TypeVar


class LogStr:
    func_args = ''


_logger = logging.getLogger(__name__)

T = TypeVar('T')

def patchbay_api(func: T) -> T:
    '''decorator for API callable functions.
    It makes debug logs and also a global logging string
    usable directly in the functions'''

    def wrapper(*args, **kwargs):
        args_strs = [str(arg) for arg in args]
        args_strs += [f"{k}={v}" for k, v in kwargs.items()]

        LogStr.func_args = f"{func.__name__}({', '.join(args_strs)})" # type:ignore
        _logger.debug(LogStr.func_args)
        return func(*args, **kwargs) # type:ignore
    return wrapper # type:ignore




import logging

from qtpy.QtCore import QPointF, QRectF, QTimer

from patshared import GroupPos, PortMode

from .init_values import canvas, options, Joining
from . import grid
from .box_widget import BoxWidget
from .api_log import LogStr

_logger = logging.getLogger(__name__)


def split_group(group_id: int, on_place=False, redraw=True):
    '''Split inputs and outputs in two box widgets.

    on_place: the new boxes will have a pos near from the existing one

    redraw: draw the box, quite long operation. Needed for 'on_place'
    to be effective.'''
    canvas.ensure_init()

    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"{LogStr.func_args} - unable to find group to split")
        return

    if group.splitted:
        _logger.error(
            f"{LogStr.func_args} - group is already splitted")
        return

    if not group.widgets:
        _logger.error(
            f"{LogStr.func_args} - group has no box widget to split")
        return

    box = group.widgets[0]
    wrap = box.is_wrapped
    ex_rect = QRectF(box.sceneBoundingRect())
    new_box = BoxWidget(group, PortMode.INPUT)
    new_box.setPos(box.pos())
    new_box.set_wrapped(wrap, animate=False)

    for portgroup in canvas.list_portgroups(group_id):
        if (portgroup.port_mode is PortMode.INPUT
                and portgroup.widget is not None):
            portgroup.widget.setParentItem(new_box)

    for port in canvas.list_ports(group_id):
        if (port.port_mode is PortMode.INPUT
                and port.widget is not None):
            port.widget.setParentItem(new_box)

    box.set_port_mode(PortMode.OUTPUT)
    group.widgets.append(new_box)
    canvas.add_box(new_box)
    group.splitted = True

    group.gpos.set_splitted(True)
    canvas.cb.group_splitted(group_id)

    if not redraw:
        return

    full_width = canvas.theme.box_spacing

    for box in group.widgets:
        box.update_positions(even_animated=True, scene_checks=False)
        full_width += box.boundingRect().width()

    if on_place:
        for box in group.widgets:
            if box.current_port_mode is PortMode.OUTPUT:
                group.gpos.boxes[PortMode.OUTPUT].pos = (
                    grid.previous_left(
                        int(ex_rect.right() + (full_width - ex_rect.width()) / 2
                            - box.boundingRect().width())),
                    grid.previous_top(
                        int(ex_rect.y()))
                )
            else:
                group.gpos.boxes[PortMode.INPUT].pos = (
                    grid.previous_left(
                        int(ex_rect.left() - (full_width - ex_rect.width()) / 2)),
                    grid.previous_top(int(ex_rect.y()))
                )

        move_group_boxes(group_id, group.gpos)
        canvas.scene.deplace_boxes_from_repulsers(
            [b for b in group.widgets if b.isVisible()])

    QTimer.singleShot(0, canvas.scene.update)

def move_group_boxes(
        group_id: int, gpos: GroupPos,
        redraw=PortMode.NULL, restore=PortMode.NULL):
    '''Highly optimized function used at view change.
    Only things that need to be redrawn are redrawn.
    Any change in this function can easily create unwanted bugs ;)

    restore is required because the previous box_pos can be hidden
    and this one shown, but without ports
    (e.g. a pure audio group in midi view)'''
    canvas.ensure_init()
    group = canvas.get_group(group_id)
    if group is None:
        return

    group.gpos = gpos
    split = gpos.is_splitted()
    join = False
    splitted = False
    orig_rect = QRectF()

    if group.splitted != split:
        if split:
            for box in group.widgets:
                if box._port_mode is PortMode.BOTH:
                    orig_rect = QRectF(box.sceneBoundingRect())
                    break

            split_group(group_id, redraw=False)
            splitted = True
            redraw |= PortMode.BOTH
        else:
            join = True

    for port_mode, box_pos, in gpos.boxes.items():
        for box in group.widgets:
            if box.port_mode is not port_mode:
                continue

            if box._layout_mode is not box_pos.layout_mode:
                box.set_layout_mode(box_pos.layout_mode)
                redraw |= port_mode

            if box.is_hidding_or_restore and not box_pos.is_hidden():
                redraw |= port_mode

            if join:
                wanted_wrap = gpos.boxes[PortMode.BOTH].is_wrapped()
            else:
                wanted_wrap = box_pos.is_wrapped()

            if box.is_wrapped is not wanted_wrap:
                # we need to update the box now, because the port_list
                # of the box is not re-evaluted when we update positions
                # during the wrap/unwrap animation.
                box.update_positions(
                    even_animated=True, scene_checks=False)
                box.set_wrapped(
                    wanted_wrap, prevent_overlap=False)
                redraw &= ~port_mode

            if redraw & port_mode:
                box.update_positions(
                    even_animated=True, scene_checks=False)

            if splitted and not orig_rect.isNull():
                # the splitted boxes start with inputs aligned to the inputs
                # of the previous joined box, and same for the outputs.
                if port_mode is PortMode.INPUT:
                    box.set_top_left((orig_rect.left(), orig_rect.top()))
                elif port_mode is PortMode.OUTPUT:
                    box.set_top_left(
                        (orig_rect.right() - box.boundingRect().width(),
                         orig_rect.top()))

            xy = grid.nearest(box_pos.pos)

            if box_pos.is_hidden():
                if box.isVisible():
                    canvas.scene.add_box_to_animation_hidding(box)

            elif restore & port_mode:
                if join:
                    canvas.scene.add_box_to_animation_restore(box)

                    both_pos = grid.nearest(gpos.boxes[PortMode.BOTH].pos)

                    if port_mode is PortMode.OUTPUT:
                        canvas.qobject.add_group_to_join(group.group_id)
                        joined_widget = BoxWidget(group, PortMode.BOTH)
                        joined_rect = joined_widget.get_dummy_rect()
                        canvas.scene.remove_box(joined_widget)
                        joined_rect.translate(QPointF(*both_pos))

                        canvas.scene.add_box_to_animation(
                            box, *both_pos,
                            joining=Joining.YES,
                            joined_rect=joined_rect)
                    else:
                        canvas.scene.add_box_to_animation(
                            box, *both_pos,
                            joining=Joining.YES)
                else:
                    box.set_top_left(xy)
                    canvas.scene.add_box_to_animation(box, *xy)
                    canvas.scene.add_box_to_animation_restore(box)

            else:
                if box.hidder_widget is not None:
                    canvas.scene.removeItem(box.hidder_widget)
                    box.hidder_widget = None

                if join:
                    both_pos = grid.nearest(gpos.boxes[PortMode.BOTH].pos)

                    if port_mode is PortMode.OUTPUT:
                        canvas.qobject.add_group_to_join(group.group_id)

                        joined_widget = BoxWidget(group, PortMode.BOTH)
                        joined_rect = joined_widget.get_dummy_rect()
                        canvas.scene.remove_box(joined_widget)
                        joined_rect.translate(QPointF(*both_pos))

                        canvas.scene.add_box_to_animation(
                            box, *both_pos,
                            joining=Joining.YES,
                            joined_rect=joined_rect)
                    else:
                        canvas.scene.add_box_to_animation(
                            box, *both_pos,
                            joining=Joining.YES)
                else:
                    canvas.scene.add_box_to_animation(
                        box, *xy, joining=Joining.NO)

def repulse_all_boxes():
    canvas.ensure_init()
    if options.prevent_overlap:
        canvas.scene.full_repulse()


from enum import Enum, auto

class Margin:
    top = 0
    'The top margin in a box with title on top'
    
    bottom = 0
    'The bottom margin in a box with title on top'
    
    sides = 0
    'The sides margin in a box with title on top'

    ports_side = 0
    'The margin on the ports side in a box with title on side'
    
    free_side = 0
    'The margin on the opposite side of the ports in a box with title on side'
    
    top_side = 0
    'The top (or the bottom) margin in a box with title on side'
    
    def __add__(self, other: 'Margin | int | float') -> 'Margin':
        new = Margin()
        if isinstance(other, Margin):
            new.top = self.top + other.top
            new.bottom = self.bottom + other.bottom
            new.sides = self.sides + other.sides
            new.ports_side = self.ports_side + other.ports_side
            new.free_side = self.free_side + other.free_side
            new.top_side = self.top_side + other.top_side
        else:
            new.top = self.top + int(other)
            new.bottom = self.bottom + int(other)
            new.sides = self.sides + int(other)
            new.ports_side = self.ports_side + int(other)
            new.free_side = self.free_side + int(other)
            new.top_side = self.top_side + int(other)
        return new

    @property
    def height(self) -> int:
        return self.top + self.bottom
    
    @property
    def width(self) -> int:
        return self.sides * 2
    
    @property
    def sided_width(self) -> int:
        return self.ports_side + self.free_side
    
    @property
    def sided_height(self) -> int:
        return self.top_side * 2

    def super(self, other: 'Margin') -> 'SuperMargin':
        return SuperMargin(self, other)


class SuperMargin(Margin):
    '''Combinaison of margins with the max values margins.
    Useful to calculate the needed sizes of a box, depending
    on the selected state, for example.'''
    def __init__(self, *margins: Margin):
        self.top = max([m.top for m in margins])
        self.bottom = max([m.bottom for m in margins])
        self.sides = max([m.sides for m in margins])
        self.ports_side = max([m.ports_side for m in margins])
        self.free_side = max([m.free_side for m in margins])
        self.top_side = max([m.top_side for m in margins])
        self._height = max([m.height for m in margins])
        self._sided_width = max([m.sided_width for m in margins])
        
    @property
    def height(self) -> int:
        return self._height
    
    @property
    def sided_width(self) -> int:
        return self._sided_width


class BorderMode(Enum):
    DEFAULT = auto()
    MINIMAL = auto()
    SIDES = auto()
    

class Align(Enum):
    LEFT = auto()
    RIGHT = auto()


import logging
from math import ceil
from struct import pack
from typing import TYPE_CHECKING

from qtpy.QtCore import Qt, QRectF, QPointF
from qtpy.QtGui import (
    QColor, QPen, QPainter, QBrush, QLinearGradient,
    QFontMetrics, QImage, QPolygonF)
# from sip import voidptr

from patshared import PortMode

from ..init_values import InlineDisplay, options, MAX_PLUGIN_ID_ALLOWED
from ..patchcanvas import canvas
from ..theme import StyleAttributer, BorderMode
from ..utils import polyline, qpolyline

from .box_utils import BoxStyler, PaintElement, UnwrapButton, WrappingState

if TYPE_CHECKING:
    from .box_widget import BoxWidget


_logger = logging.getLogger(__name__)


def _paint_ports_border_lines(box: 'BoxWidget', painter: QPainter):
    theme = box.get_theme(BoxStyler.PORTS_BORDER)
    if not theme.visible:
        return

    box_theme = box.get_theme()
    border_width = box_theme.border_width
    painter.setPen(theme.fill_pen)    
    lh = theme.border_width / 2

    for port_mode in (PortMode.OUTPUT, PortMode.INPUT):
        if not box.current_port_mode & port_mode:
            continue

        if port_mode is PortMode.INPUT:
            x = border_width + lh
        else:
            x = box._width - border_width - lh
        
        if box.has_side_title:
            painter.drawLine(
                QPointF(x, border_width + lh),
                QPointF(x, box._height - border_width -lh))

        else:
            header_theme = box.get_theme(BoxStyler.HEADER)
            painter.drawLine(
                QPointF(x,
                        border_width + box._header_height
                            - header_theme.margin.bottom + lh),
                QPointF(x, box._height - border_width - lh))

def _paint_hardware_rack(box: 'BoxWidget', painter: QPainter):
    if not box.is_hardware:
        return

    if box._layout is None:
        return

    d = float(canvas.theme.hardware_rack_width)
    sd = d * 0.5

    theme = canvas.theme.hardware_rack
    if box.isSelected():
        theme = theme.selected

    background1 = theme.background_color
    background2 = theme.background2_color

    if background2 is not None:
        hw_gradient = QLinearGradient(
            -d, -d, box._width + d, box._height + d)
        hw_gradient.setColorAt(0, background1)
        hw_gradient.setColorAt(0.5, background2)
        hw_gradient.setColorAt(1, background1)

        painter.setBrush(hw_gradient)
    else:
        painter.setBrush(background1)

    pen = theme.fill_pen
    painter.setPen(pen)
    lh = pen.widthF() / 2.0

    ports_top_in = box._layout.ports_top_in
    ports_top_out = box._layout.ports_top_out
    ports_bottom_in = box._layout.ports_bottom_in
    ports_bottom_out = box._layout.ports_bottom_out

    if box.current_port_mode is not PortMode.BOTH:
        if box.current_port_mode is PortMode.INPUT:
            points = [
                (- lh, - lh),
                (- lh, ports_top_in - lh),
                (- sd, ports_top_in - lh),
                (- d + lh, ports_top_in - sd),
                (- d + lh, - sd),
                (- sd, - d + lh),
                (box._width + sd, - d + lh),
                (box._width + d - lh, -sd),
                (box._width + d - lh, box._height - lh + sd),
                (box._width + sd, box._height + d - lh),
                (- sd, box._height + d - lh),
                (-d + lh, box._height + sd),
                (-d + lh, ports_bottom_in + sd),
                (- sd, ports_bottom_in + lh),
                (- lh, ports_bottom_in + lh),
                (- lh, box._height + lh),
                (box._width + lh, box._height + lh),
                (box._width + lh, - lh)
            ]

        else:
            points = [
                (box._width + lh, - lh),
                (box._width + lh, ports_top_out - lh),
                (box._width + sd, ports_top_out - lh),
                (box._width + d - lh, ports_top_out - sd),
                (box._width + d - lh, - sd),
                (box._width + sd, -d + lh),
                (- sd, -d + lh),
                (-d + lh, - sd),
                (-d + lh, box._height + sd),
                (- sd, box._height + d - lh),
                (box._width + sd, box._height + d - lh),
                (box._width + d - lh, box._height + sd),
                (box._width + d - lh, ports_bottom_out + sd),
                (box._width + sd, ports_bottom_out + lh),
                (box._width + lh, ports_bottom_out + lh),
                (box._width + lh, box._height + lh),
                (-lh, box._height + lh),
                (-lh, -lh)
            ]

        hardware_poly = QPolygonF()
        for xy in points:
            hardware_poly += QPointF(*xy)

        if theme.border_mode is BorderMode.MINIMAL:
            painter.drawPolyline(polyline(points[2:6]))
            painter.drawPolyline(polyline(points[10:14]))
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawPolygon(hardware_poly)
        else:
            painter.drawPolygon(hardware_poly)
    else:
        top_points = [
            (- lh, - lh),
            (- lh, ports_top_in - lh),
            (- sd, ports_top_in - lh),
            (- d + lh, ports_top_in - sd),
            (- d + lh, - sd),
            (- sd, -d + lh),
            (box._width + sd, -d + lh),
            (box._width + d - lh, - sd),
            (box._width + d - lh, ports_top_out - sd),
            (box._width + d/2, ports_top_out - lh),
            (box._width + lh, ports_top_out - lh),
            (box._width + lh, -lh)
        ]

        bottom_points = [
            (- lh, box._height + lh),
            (- lh, ports_bottom_in + lh),
            (- sd, ports_bottom_in + lh),
            (- d + lh, ports_bottom_in + sd),
            (-d + lh, box._height + sd),
            (- sd, box._height + d - lh),
            (box._width + sd, box._height + d - lh),
            (box._width + d - lh, box._height + sd),
            (box._width + d - lh, ports_bottom_out + sd),
            (box._width + sd, ports_bottom_out + lh),
            (box._width + lh, ports_bottom_out + lh),
            (box._width + lh, box._height + lh)
        ]

        hw_poly_top = QPolygonF()
        for xy in top_points:
            hw_poly_top += QPointF(*xy)

        hw_poly_bottom = QPolygonF()
        for xy in bottom_points:
            hw_poly_bottom += QPointF(*xy)

        if theme.border_mode is BorderMode.MINIMAL:
            painter.drawPolyline(polyline(top_points[2:6]))
            painter.drawPolyline(polyline(top_points[6:10]))
            painter.drawPolyline(polyline(bottom_points[2:6]))
            painter.drawPolyline(polyline(bottom_points[6:10]))
            painter.setPen(Qt.PenStyle.NoPen)
            painter.drawPolygon(hw_poly_top)
            painter.drawPolygon(hw_poly_bottom)
        else:
            painter.drawPolygon(hw_poly_top)
            painter.drawPolygon(hw_poly_bottom)

def _paint_inline_display(box: 'BoxWidget', painter: QPainter):
    if box._plugin_inline is InlineDisplay.DISABLED:
        return
    if not options.inline_displays:
        return

    inwidth  = box._width - box._width_in - box._width_out - 16
    inheight = (box._height - box._header_height
                - box.get_theme().port_spacing - 3)
    scaling  = (canvas.scene.get_scale_factor()
                * canvas.scene.get_device_pixel_ratio_f())

    if (box._plugin_id >= 0
            and box._plugin_id <= MAX_PLUGIN_ID_ALLOWED
            and (box._plugin_inline is InlineDisplay.ENABLED
                    or box._inline_scaling != scaling)):
        data = canvas.cb.inline_display(
            box._plugin_id,
            int(inwidth*scaling), int(inheight*scaling))

        if data is None:
            return

        # invalidate old image first
        del box._inline_image

        box._inline_data = pack(
            "%iB" % (data['height'] * data['stride']), *data['data'])
        box._inline_image = QImage(
            voidptr(box._inline_data), data['width'], data['height'], # type:ignore
            data['stride'], QImage.Format.Format_ARGB32)
        box._inline_scaling = scaling
        box._plugin_inline = InlineDisplay.CACHED

    if box._inline_image is None:
        _logger.warning(
            'inline display image is None for '
            f'{box._plugin_id}, {box._group_name}')
        return

    swidth = box._inline_image.width() / scaling
    sheight = box._inline_image.height() / scaling

    srcx = int(box._width_in
                + (box._width - box._width_in - box._width_out) / 2
                - swidth / 2)
    srcy = int(box._header_height + 1 + (inheight - sheight) / 2)

    painter.drawImage(
        QRectF(srcx, srcy, swidth, sheight), box._inline_image)

def _paint_gui_button(box: 'BoxWidget', painter: QPainter, border: int):
    gui_theme = canvas.theme.gui_button
    if box._gui_visible:
        gui_theme = gui_theme.gui_visible
    else:
        gui_theme = gui_theme.gui_hidden            
    
    gmg = gui_theme.margin
    header_theme = box.get_theme(BoxStyler.HEADER)
    selected = box.isSelected()
    if selected:
        header_theme = header_theme.selected
    hmg = header_theme.margin
    mg = hmg + gmg
    gui_lh = gui_theme.border_width / 2
    
    if box.has_side_title:
        if box.current_port_mode is PortMode.INPUT:
            gui_rect = QRectF(
                box._width - box._header_width - border + mg.ports_side + gui_lh,
                mg.top_side + border + gui_lh,
                box._header_width - mg.sided_width - gui_lh * 2,
                box._header_height - mg.sided_height - gui_lh * 2)
        else:
            gui_rect = QRectF(
                border + mg.free_side + gui_lh,
                border + mg.top_side + gui_lh,
                box._header_width - mg.sided_width - 2 * gui_lh,
                box._header_height - mg.sided_height - 2 * gui_lh)
    else:
        space_left = (box._width - box._header_width - 2 * border) / 2
        gui_rect = QRectF(
            border + mg.sides + space_left + gui_lh,
            border + mg.top + gui_lh,
            box._header_width - mg.width - gui_lh * 2,
            box._header_height - mg.height - gui_lh * 2)

    painter.setBrush(
        _get_gradient(gui_theme.background_color,
                      gui_theme.background2_color,
                      box._header_width, box._header_height))
    painter.setPen(gui_theme.fill_pen)
    
    match gui_theme.border_mode:            
        case BorderMode.MINIMAL:
            painter.drawPolyline(qpolyline(
                gui_rect.bottomLeft(),
                gui_rect.topLeft(),
                gui_rect.topRight(),
                gui_rect.bottomRight()))
            painter.setPen(Qt.PenStyle.NoPen)
        case BorderMode.SIDES:
            painter.drawPolyline(qpolyline(
                gui_rect.bottomLeft(), gui_rect.topLeft()))
            painter.drawPolyline(qpolyline(
                gui_rect.topRight(), gui_rect.bottomRight()))
            painter.setPen(Qt.PenStyle.NoPen)

    sel_paths = box._painter_paths.get(selected)
    if sel_paths is not None:
        gui_path = sel_paths.get(PaintElement.GUI_BUTTON)
        if gui_path is not None:
            painter.drawPath(gui_path)

def _paint_monitor_deco(box: 'BoxWidget', painter: QPainter, pen_width: int):
    if box.current_port_mode is PortMode.OUTPUT:
        bor_gradient = QLinearGradient(0, 0, box._height, box._height)
    else:
        bor_gradient = QLinearGradient(
            box._width, 0, box._height, box._width - box._height)

    mon_theme = canvas.theme.monitor_decoration
    if box.isSelected():
        mon_theme = mon_theme.selected

    color_main = mon_theme.background_color
    color_alter = mon_theme.background2_color

    if color_alter is not None:
        tot = int(box._height / 20)
        for i in range(tot):
            if i % 2 == 0:
                bor_gradient.setColorAt(i/tot, color_main)
            else:
                bor_gradient.setColorAt(i/tot, color_alter)

        painter.setBrush(bor_gradient)
    else:
        painter.setBrush(color_main)

    BAND_MON_WIDTH = 9
    TRIANGLE_MON_SIZE_TOP = 7
    triangle_mon_size_bottom = 0
    if (box._wrapping_state in (WrappingState.WRAPPING,
                                    WrappingState.UNWRAPPING)
            or (box._wrapping_state is WrappingState.NORMAL
                and box._wrap_triangle_pos is not UnwrapButton.NONE)):
        triangle_mon_size_bottom = 13

    bmw = BAND_MON_WIDTH
    tms_top = TRIANGLE_MON_SIZE_TOP
    tms_bot = triangle_mon_size_bottom

    xside = pen_width
    xband = pen_width + bmw
    xtop = pen_width + bmw + tms_top
    xbot = pen_width + bmw + tms_bot

    if box.current_port_mode is PortMode.INPUT:
        xside = box._width - xside
        xband = box._width - xband
        xtop = box._width - xtop
        xbot = box._width - xbot

    mon_points = [
        (xside, pen_width),
        (xtop, pen_width),
        (xband, pen_width + tms_top),
        (xband, box._height - tms_bot - pen_width),
        (xbot, box._height - pen_width),
        (xside, box._height - pen_width)]

    mon_poly = QPolygonF()
    for xy in mon_points:
        mon_poly += QPointF(*xy)
    
    if mon_theme.border_mode is BorderMode.MINIMAL:
        painter.setPen(Qt.PenStyle.NoPen)
        painter.drawPolygon(mon_poly)
        painter.setPen(mon_theme.fill_pen)
        painter.drawPolyline(polyline(mon_points[1:5]))
    else:
        painter.setPen(mon_theme.fill_pen)
        painter.drawPolygon(mon_poly)

def _paint_header_lines(box: 'BoxWidget', painter: QPainter):
    ...
    
def _paint_title_lines(box: 'BoxWidget', painter: QPainter,
                       normal_color: QColor):
    if box._can_handle_gui:
        # allow text color to be defined in gui_button theme entry
        if box._gui_visible:
            gui_text_color = canvas.theme.gui_button.gui_visible.text_color
        else:
            gui_text_color = canvas.theme.gui_button.gui_hidden.text_color
        
        if gui_text_color.isValid():
            normal_color = gui_text_color

    opac_color = QColor(normal_color)
    opac_color.setAlpha(int(normal_color.alpha() / 2))

    text_pen = QPen(normal_color)
    opac_text_pen = QPen(opac_color)

    # draw title lines
    for title_line in box._title_lines:
        painter.setFont(title_line.get_font())

        if title_line.is_little:
            painter.setPen(opac_text_pen)
        else:
            painter.setPen(text_pen)

        if (box.is_monitor
                and title_line == box._title_lines[-1]
                and box._group_name.endswith(' Monitor')):
            # Title line endswith " Monitor"
            # Draw "Monitor" in yellow
            # but keep the rest in white
            pre_text = title_line.text.rpartition(' Monitor')[0]
            painter.drawText(
                ceil(title_line.x), ceil(title_line.y), pre_text)

            x_pos = title_line.x
            if pre_text:
                t_font = title_line.get_font()
                x_pos += QFontMetrics(t_font).horizontalAdvance(pre_text)
                x_pos += QFontMetrics(t_font).horizontalAdvance(' ')

            painter.setPen(QPen(canvas.theme.monitor_color, 0))
            painter.drawText(ceil(x_pos), ceil(title_line.y), 'Monitor')
        else:
            painter.drawText(
                ceil(title_line.x), ceil(title_line.y), title_line.text)

def _paint_wrappers(
        box: 'BoxWidget', painter: QPainter, wtheme: StyleAttributer,
        pen_width: float):
    painter.setPen(wtheme.fill_pen)
    painter.setBrush(wtheme.background_color)

    match box._wrap_triangle_pos:
        case _ if box._wrapping_state in(
                WrappingState.WRAPPED, WrappingState.UNWRAPPING):
            for port_mode in PortMode.INPUT, PortMode.OUTPUT:
                if not box.current_port_mode & port_mode:
                    continue
                
                painter.setPen(wtheme.fill_pen)
                
                if box.has_side_title:
                    side = 8.5
                    ypos = box._height - pen_width - 2.0

                    triangle = QPolygonF()
                    if port_mode is PortMode.INPUT:
                        xpos = pen_width + 2.0
                        triangle += QPointF(xpos, ypos - side)
                        triangle += QPointF(xpos + side, ypos)
                        triangle += QPointF(xpos, ypos)
                    else:
                        xpos = box._width - pen_width - 2.0
                        triangle += QPointF(xpos, ypos - side)
                        triangle += QPointF(xpos - side, ypos)
                        triangle += QPointF(xpos, ypos)
                else:
                    side = 6
                    xpos = pen_width + 2.0
                    ypos = box._height - pen_width - side - 2.0

                    if port_mode is PortMode.OUTPUT:
                        xpos = \
                            box._width - pen_width - 2.0 - 2 * side

                    triangle = QPolygonF()
                    triangle += QPointF(xpos, ypos)
                    triangle += QPointF(xpos + 2 * side, ypos)
                    triangle += QPointF(xpos + side, ypos + side)

                if wtheme.border_mode is BorderMode.MINIMAL:
                    painter.drawPolyline(qpolyline(
                        triangle[0], triangle[2], triangle[1]))
                    painter.setPen(Qt.PenStyle.NoPen)
                painter.drawPolygon(triangle)

        case UnwrapButton.LEFT:
            side = 6
            xpos = 2.0 + pen_width
            ypos = box._height - pen_width - 2.0
            triangle = QPolygonF()
            triangle += QPointF(xpos, ypos)
            triangle += QPointF(xpos + 2 * side, ypos)
            triangle += QPointF(xpos + side, ypos -side)

            if wtheme.border_mode is BorderMode.MINIMAL:
                painter.drawPolyline(qpolyline(
                    triangle[0], triangle[2], triangle[1]))
                painter.setPen(Qt.PenStyle.NoPen)
            painter.drawPolygon(triangle)

        case UnwrapButton.RIGHT:
            side = 6
            xpos = box._width - pen_width - 2 * side - 2.0

            ypos = box._height - pen_width - 2.0
            triangle = QPolygonF()
            triangle += QPointF(xpos, ypos)
            triangle += QPointF(xpos + 2 * side, ypos)
            triangle += QPointF(xpos + side, ypos - side)
            
            if wtheme.border_mode is BorderMode.MINIMAL:
                painter.drawPolyline(qpolyline(
                    triangle[0], triangle[2], triangle[1]))
                painter.setPen(Qt.PenStyle.NoPen)
            painter.drawPolygon(triangle)

        case UnwrapButton.CENTER:
            side = 7
            pms = box.get_layout()._pms
            xpos = (box._width + pms.ins_width - pms.outs_width) / 2 - side

            ypos = box._height - pen_width / 2.0
            triangle = QPolygonF()
            triangle += QPointF(xpos, ypos)
            triangle += QPointF(xpos + 2 * side, ypos)
            triangle += QPointF(xpos + side, ypos -side)
            
            if wtheme.border_mode is BorderMode.MINIMAL:
                painter.drawPolyline(qpolyline(
                    triangle[0], triangle[2], triangle[1]))
                painter.setPen(Qt.PenStyle.NoPen)
            painter.drawPolygon(triangle)

def _get_gradient(
        color_main: QColor, color_alter: QColor | None,
        width: int, height: int) -> QLinearGradient | QColor:
    if color_alter is None:
        return color_main
    
    max_size = max(width, height)
    box_gradient = QLinearGradient(0, 0, max_size, max_size)
    gradient_size = 20

    box_gradient.setColorAt(0, color_main)
    tot = int(max_size / gradient_size)
    for i in range(tot):
        if i % 2 == 0:
            box_gradient.setColorAt((i/tot) ** 0.7, color_main)
        else:
            box_gradient.setColorAt((i/tot) ** 0.7, color_alter)

    return box_gradient

def paint(box: 'BoxWidget', painter: QPainter, option, widget):
    if canvas.loading_items:
        return

    if box._layout is None:
        return

    painter.save()
    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)

    # define theme for box, wrappers and header lines
    theme = box.get_theme()
    wtheme = box.get_theme(BoxStyler.WRAPPER)
    htheme = box.get_theme(BoxStyler.HEADER)
    hltheme = box.get_theme(BoxStyler.HEADER_LINE)

    border_unselected = int(theme.border_width)
    selected = box.isSelected()
    
    if selected:
        theme = theme.selected
        wtheme = wtheme.selected
        hltheme = hltheme.selected

    bg_image = theme.background_image
    painter_paths = box._painter_paths.get(selected)
    if painter_paths is None:
        _logger.error(f'Ask to paint {box} but no QPainterPath is created')
        return

    main_ppath = painter_paths.get(PaintElement.MAIN)
    header_ppath = painter_paths.get(PaintElement.HEADER)
    anti_header_ppath = painter_paths.get(PaintElement.ANTI_HEADER)

    # draw the background image if exists
    if not bg_image.isNull():
        painter.setBrush(QBrush(bg_image))
        painter.setPen(Qt.PenStyle.NoPen)
        if anti_header_ppath is not None:
            painter.drawPath(anti_header_ppath)
        elif main_ppath is not None:
            painter.drawPath(main_ppath)

    # draw the main rectangle
    pen = theme.fill_pen
    painter.setPen(pen)
    pen_width = theme.border_width

    painter.setBrush(
        _get_gradient(
            theme.background_color, theme.background2_color,
            box._width, box._height))
    
    painter.setPen(Qt.PenStyle.NoPen)
    
    if anti_header_ppath is not None:
        painter.drawPath(anti_header_ppath)
    elif main_ppath is not None:
        # should not happen
        painter.drawPath(main_ppath)

    _paint_ports_border_lines(box, painter)

    if header_ppath is not None:
        hbg_image = htheme.background_image
        if not hbg_image.isNull():
            painter.setBrush(QBrush(hbg_image))
            painter.drawPath(header_ppath)

        painter.setPen(htheme.fill_pen)
        painter.setBrush(_get_gradient(
            htheme.background_color, htheme.background2_color,
            box._width, box._height))
        painter.drawPath(header_ppath)

    painter.setPen(pen)        
    painter.setBrush(Qt.BrushStyle.NoBrush)
    if main_ppath is not None:
        painter.drawPath(main_ppath)

    # draw hardware box decoration (flyrack like)
    _paint_hardware_rack(box, painter)

    # Draw plugin inline display if supported
    _paint_inline_display(box, painter)

    # Draw toggle GUI client button
    if box._can_handle_gui:
        _paint_gui_button(box, painter, border_unselected)

    # draw Pipewire Monitor (or PulseAudio bridges) decorations
    elif box.is_monitor and box.current_port_mode is not PortMode.BOTH:
        _paint_monitor_deco(box, painter, int(pen_width))

    # may draw horizontal lines around title (header lines)
    if (box._header_line_left is not None
            and box._header_line_right is not None):
        painter.setPen(hltheme.fill_pen)
        painter.drawLine(QPointF(*box._header_line_left[0:2]),
                         QPointF(*box._header_line_left[2:]))
        painter.drawLine(QPointF(*box._header_line_right[0:2]),
                         QPointF(*box._header_line_right[2:]))

    _paint_title_lines(box, painter, theme.text_color)

    # draw (un)wrapper triangles
    _paint_wrappers(box, painter, wtheme, pen_width)

    painter.restore()

